@@ -8,13 +8,16 @@
 -- General character property related functions.
 --
 module Unicode.Char.General
-    (
-    -- * Unicode general categories
-      GeneralCategory(..)
+    ( -- * Types of Code Points
+      CodePointType(..)
+    , codePointType
+
+      -- * Unicode general categories
+    , GeneralCategory(..)
     , generalCategoryAbbr
     , generalCategory
 
-    -- * Character classification
+      -- * Character classification
     , isAlphabetic
     , isAlphaNum
     , isControl
@@ -24,9 +27,9 @@ module Unicode.Char.General
     , isSeparator
     , isSymbol
     , isWhiteSpace
-    , isLetter
-    , isSpace
-    -- ** Re-export
+    , isNoncharacter
+
+      -- ** Re-export
     , isAscii
     , isLatin1
     , isAsciiUpper
@@ -94,19 +97,25 @@ module Unicode.Char.General
 where
 
 import Control.Exception (assert)
+import Data.Bits ((.&.))
 import Data.Char (isAscii, isLatin1, isAsciiUpper, isAsciiLower, ord)
 import Data.Ix (Ix)
-import Unicode.Internal.Division (quotRem28)
 
+import qualified Unicode.Char.General.Compat as Compat
 import qualified Unicode.Internal.Char.DerivedCoreProperties as P
 import qualified Unicode.Internal.Char.PropList as P
 import qualified Unicode.Internal.Char.UnicodeData.GeneralCategory as UC
+import Unicode.Internal.Division (quotRem28)
+
+--------------------------------------------------------------------------------
+-- General Category
+--------------------------------------------------------------------------------
 
 {-| Unicode General Categories.
 
 These classes are defined in the
 [Unicode Character Database](http://www.unicode.org/reports/tr44/tr44-14.html#GC_Values_Table),
-part of the Unicode standard
+part of the Unicode standard.
 
 __Note:__ the classes must be in the same order they are listed in the Unicode Standard,
 because some functions (e.g. 'generalCategory') rely on the 'Enum' instance.
@@ -216,6 +225,116 @@ prop> show (generalCategory c) == show (Data.Char.generalCategory c)
 generalCategory :: Char -> GeneralCategory
 generalCategory = toEnum . UC.generalCategory
 
+--------------------------------------------------------------------------------
+-- Types of Code Points
+--------------------------------------------------------------------------------
+
+-- | Types of Code Points.
+--
+-- These classes are defined in the section
+-- [2.4 “Code Points and Characters”](https://www.unicode.org/versions/Unicode15.0.0/ch02.pdf#G14527)
+-- of the Unicode standard.
+--
+-- @since 0.4.1
+data CodePointType
+    = GraphicType
+    -- ^ __Graphic__: defined by the following general categories:
+    --
+    -- * Letters (L): 'UppercaseLetter', 'LowercaseLetter', 'TitlecaseLetter',
+    --   'ModifierLetter', 'OtherLetter'.
+    -- * Marks (M): 'NonSpacingMark', 'SpacingCombiningMark', 'EnclosingMark'.
+    -- * Numbers (N): 'DecimalNumber', 'LetterNumber', 'OtherNumber'.
+    -- * Punctuation (P): 'ConnectorPunctuation', 'DashPunctuation',
+    --   'OpenPunctuation', 'ClosePunctuation', 'InitialQuote', 'FinalQuote',
+    --   'OtherPunctuation'.
+    -- * Symbol (S): 'MathSymbol', 'CurrencySymbol', 'ModifierSymbol',
+    --   'OtherSymbol'.
+    -- * Separators: 'Space'.
+    | FormatType
+    -- ^ __Format__: invisible but affects neighboring characters.
+    --
+    -- Defined by the following general categories:
+    -- 'LineSeparator', 'ParagraphSeparator', 'Format'.
+    | ControlType
+    -- ^ __Control__: usage defined by protocols or standards outside the
+    -- Unicode Standard.
+    --
+    -- Defined by the general category 'Control'.
+    | PrivateUseType
+    -- ^ __Private-use__: usage defined by private agreement outside the
+    -- Unicode Standard.
+    --
+    -- Defined by the general category 'PrivateUse'.
+    | SurrogateType
+    -- ^ __Surrogate__: Permanently reserved for UTF-16.
+    --
+    -- Defined by the general category 'Surrogate'.
+    | NoncharacterType
+    -- ^ __Noncharacter:__ a code point that is permanently reserved for
+    -- internal use (see definition D14 in the section
+    -- [3.4 “Characters and Encoding”](https://www.unicode.org/versions/Unicode15.0.0/ch03.pdf#G2212)
+    -- of the Unicode Standard).
+    -- Noncharacters consist of the values @U+nFFFE@ and @U+nFFFF@ (where @n@
+    -- is from 0 to 10₁₆) and the values @U+FDD0..U+FDEF@.
+    --
+    -- They are a subset of the general category 'NotAssigned'.
+    | ReservedType
+    -- ^ __Reserved:__ any code point of the Unicode Standard that is reserved
+    -- for future assignment (see definition D15 in the section
+    -- [3.4 “Characters and Encoding”](https://www.unicode.org/versions/Unicode15.0.0/ch03.pdf#G2212)
+    -- of the Unicode Standard). Also known as an unassigned code point.
+    --
+    -- They are a subset of the general category 'NotAssigned'.
+    deriving ( Show
+             , Eq
+             , Ord
+             , Enum
+             , Bounded
+             , Ix
+             )
+
+-- | Returns the 'CodePointType' of a character.
+--
+-- @since 0.6.0
+codePointType :: Char -> CodePointType
+codePointType c = case generalCategory c of
+    UppercaseLetter      -> GraphicType
+    LowercaseLetter      -> GraphicType
+    TitlecaseLetter      -> GraphicType
+    ModifierLetter       -> GraphicType
+    OtherLetter          -> GraphicType
+    NonSpacingMark       -> GraphicType
+    SpacingCombiningMark -> GraphicType
+    EnclosingMark        -> GraphicType
+    DecimalNumber        -> GraphicType
+    LetterNumber         -> GraphicType
+    OtherNumber          -> GraphicType
+    ConnectorPunctuation -> GraphicType
+    DashPunctuation      -> GraphicType
+    OpenPunctuation      -> GraphicType
+    ClosePunctuation     -> GraphicType
+    InitialQuote         -> GraphicType
+    FinalQuote           -> GraphicType
+    OtherPunctuation     -> GraphicType
+    MathSymbol           -> GraphicType
+    CurrencySymbol       -> GraphicType
+    ModifierSymbol       -> GraphicType
+    OtherSymbol          -> GraphicType
+    Space                -> GraphicType
+    LineSeparator        -> FormatType
+    ParagraphSeparator   -> FormatType
+    Control              -> ControlType
+    Format               -> FormatType
+    Surrogate            -> SurrogateType
+    PrivateUse           -> PrivateUseType
+    NotAssigned
+        | isNoncharacter c -> NoncharacterType
+        | otherwise        -> ReservedType
+
+--------------------------------------------------------------------------------
+-- Predicates
+--------------------------------------------------------------------------------
+
 {-| Returns 'True' for alphabetic Unicode characters (lower-case, upper-case
 and title-case letters, plus letters of caseless scripts and modifiers
 letters).
@@ -263,20 +382,19 @@ following 'GeneralCategory's, or 'False' otherwise:
 
 prop> isAlphaNum c == Data.Char.isAlphaNum c
 
+__Note:__ this function is incompatible with 'isAlphabetic':
+
+>>> isAlphabetic '\x345'
+True
+>>> isAlphaNum '\x345'
+False
+
 @since 0.3.0
 -}
+{-# INLINE isAlphaNum #-}
+{-# DEPRECATED isAlphaNum "Use Unicode.Char.General.Compat.isAlphaNum instead." #-}
 isAlphaNum :: Char -> Bool
-isAlphaNum c =
-    let !cp = ord c
-    -- NOTE: The guard constant is updated at each Unicode revision.
-    --       It must be < 0x40000 to be accepted by generalCategoryPlanes0To3.
-    in cp <= UC.MaxIsAlphaNum &&
-        let !gc = UC.generalCategoryPlanes0To3 cp
-        in gc <= UC.OtherLetter ||
-           (UC.DecimalNumber <= gc && gc <= UC.OtherNumber)
-    -- Use the following in case the previous code is not valid anymore:
-    -- gc <= UC.OtherLetter || (UC.DecimalNumber <= gc && gc <= UC.OtherNumber)
-    -- where !gc = UC.generalCategory c
+isAlphaNum = Compat.isAlphaNum
 
 {-| Selects control characters, which are the non-printing characters
 of the Latin-1 subset of Unicode.
@@ -414,24 +532,21 @@ isSymbol :: Char -> Bool
 isSymbol c = UC.MathSymbol <= gc && gc <= UC.OtherSymbol
     where gc = UC.generalCategory c
 
--- | Returns 'True' for alphabetic Unicode characters (lower-case, upper-case
--- and title-case letters, plus letters of caseless scripts and modifiers
--- letters).
+-- | Returns 'True' for any /noncharacter/.
 --
--- @since 0.1.0
-{-# INLINE isLetter #-}
-{-# DEPRECATED isLetter "Use isAlphabetic instead. Note that the behavior of this function does not match base:Data.Char.isLetter. See Unicode.Char.General.Compat for behavior compatible with base:Data.Char." #-}
-isLetter :: Char -> Bool
-isLetter = P.isAlphabetic
-
--- | Returns 'True' for any whitespace characters, and the control
--- characters @\\t@, @\\n@, @\\r@, @\\f@, @\\v@.
+-- A /noncharacter/ is a code point that is permanently reserved for internal
+-- use (see definition D14 in the section
+-- [3.4 “Characters and Encoding”](https://www.unicode.org/versions/Unicode15.0.0/ch03.pdf#G2212)
+-- of the Unicode Standard).
 --
--- @since 0.1.0
-{-# INLINE isSpace #-}
-{-# DEPRECATED isSpace "Use isWhiteSpace instead. Note that the behavior of this function does not match base:Data.Char.isSpace. See Unicode.Char.General.Compat for behavior compatible with base:Data.Char." #-}
-isSpace :: Char -> Bool
-isSpace = P.isWhite_Space
+-- Noncharacters consist of the values @U+nFFFE@ and @U+nFFFF@ (where @n@
+-- is from 0 to 10₁₆) and the values @U+FDD0..U+FDEF@.
+--
+-- @since 0.6.0
+isNoncharacter :: Char -> Bool
+isNoncharacter c
+    = ('\xFDD0' <= c && c <= '\xFDEF')
+    || (ord c .&. 0xFFFF) >= 0xFFFE
 
 -------------------------------------------------------------------------------
 -- Korean Hangul
